#define _WIN32_WINDOWS 0x410
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <float.h>
#include "m_pd.h"
#include "pdTabletTransfer.h"
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )

#define DEFPOLLTIME 5
#define MUTEXWAITTIME 1

static t_class *wintablet_class;

typedef struct _wintablet
{
  t_object x_ob;
  HANDLE x_transferMutex;
  HANDLE x_transferFileMap;
  PROCESS_INFORMATION x_appProcessInfo;
  pdTabletTransferData *x_wintabletData;
  t_clock *x_pollClock;
  int x_pollTime;
  t_outlet *x_outletY;
  t_outlet *x_outletPressure;
  t_outlet *x_outletOrientation;
  t_outlet *x_outletButtons;
} t_wintablet;

void wintablet_outputData(t_wintablet *x);
void wintablet_openApp(t_wintablet *x);
void wintablet_closeApp(t_wintablet *x);
int wintablet_appIsRunning(t_wintablet *x);

void *wintablet_new() {
  t_wintablet *x = (t_wintablet *)pd_new(wintablet_class);
  outlet_new(&x->x_ob, &s_float);
  x->x_outletY = outlet_new(&x->x_ob, &s_float);
  x->x_outletPressure = outlet_new(&x->x_ob, &s_float);
  x->x_outletOrientation = outlet_new(&x->x_ob, &s_float);
  x->x_outletButtons = outlet_new(&x->x_ob, &s_float);
  x->x_pollTime = DEFPOLLTIME;
  x->x_pollClock = clock_new (x, (t_method)wintablet_outputData);
  ZeroMemory(&(x->x_appProcessInfo), sizeof(x->x_appProcessInfo));
  return (x);
}

void wintablet_openApp(t_wintablet *x) {
  char mutexName[1024], fileMapName[1024];
  char cmdLineArgs[2048];
  STARTUPINFO si;

  if (wintablet_appIsRunning(x)) {
    post("wintablet: wintablet.exe is already running");
    return;
  }
  sprintf(mutexName, "mutex%x", x);
  sprintf(fileMapName, "fileMap%x", x);
  sprintf(cmdLineArgs, 
          "wintablet.exe wintablet.exe %s %s %d", 
          fileMapName, 
          mutexName,
          GetCurrentProcessId());
  x->x_transferMutex = CreateMutex(NULL, 0, mutexName);
  x->x_transferFileMap = CreateFileMapping(INVALID_HANDLE_VALUE,
                                           NULL,
                                           PAGE_READWRITE,
                                           0,
                                           sizeof(pdTabletTransferData),
                                           fileMapName);
  x->x_wintabletData = (pdTabletTransferData *)MapViewOfFile(x->x_transferFileMap,
                                                          FILE_MAP_ALL_ACCESS,
                                                          0,
                                                          0,
                                                          sizeof(pdTabletTransferData));
  ZeroMemory(&si, sizeof(si));
  si.cb = sizeof(si);
  if (CreateProcess(NULL,
                    cmdLineArgs,
                    NULL,
                    NULL,
                    0,
                    0,
                    NULL,
                    NULL,
                    &si,
                    &(x->x_appProcessInfo)) == 0) {
    char errorString[1024];

    FormatMessage(
        FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        GetLastError(),
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR)errorString,
        1024, 
        NULL);
    post("wintablet: error opening wintablet.exe");
    post("wintablet: %s", errorString);
  }
  else {
    clock_delay(x->x_pollClock, x->x_pollTime);
  }
}

void wintablet_closeApp(t_wintablet *x) {
  TerminateProcess(x->x_appProcessInfo.hProcess, 0);
  ZeroMemory(&(x->x_appProcessInfo), sizeof(x->x_appProcessInfo));
}

int wintablet_appIsRunning(t_wintablet *x) {
  int exitCode = 0;
  
  GetExitCodeProcess(x->x_appProcessInfo.hProcess, &exitCode);
  return (exitCode == STILL_ACTIVE);
}

void wintablet_outputData(t_wintablet *x) {
  int i;
  
  if (x->x_wintabletData && x->x_transferMutex) {
    if (WaitForSingleObject(x->x_transferMutex, MUTEXWAITTIME) != WAIT_TIMEOUT) {
      for (i = 0; i < x->x_wintabletData->nEvents; i++) {
        outlet_float(x->x_outletButtons, 
                     x->x_wintabletData->eventBuffer[i].buttons);
        outlet_float(x->x_outletOrientation, 
                     x->x_wintabletData->eventBuffer[i].orientation);
        outlet_float(x->x_outletPressure, 
                     x->x_wintabletData->eventBuffer[i].pressure);
        outlet_float(x->x_outletY, 
                     x->x_wintabletData->eventBuffer[i].y);
        outlet_float(x->x_ob.ob_outlet, 
                     x->x_wintabletData->eventBuffer[i].x);
      }
      x->x_wintabletData->nEvents = 0;
      ReleaseMutex(x->x_transferMutex);
    }
  }
  clock_delay(x->x_pollClock, x->x_pollTime);
}

void wintablet_free(t_wintablet *x) {
  clock_unset(x->x_pollClock);
  clock_free(x->x_pollClock);
  UnmapViewOfFile(x->x_wintabletData);
  CloseHandle(x->x_transferFileMap);
  CloseHandle(x->x_transferMutex);
  wintablet_closeApp(x);
}

void wintablet_setup(void) {
  wintablet_class = class_new(gensym("wintablet"),
                              (t_newmethod)wintablet_new,
                              (t_method)wintablet_free,
                              sizeof(t_wintablet),
                              0,
                              0);
  class_addbang(wintablet_class, (t_method)wintablet_openApp); 
}

